/*
	File:		FWList.cpp

	Contains:	Primitive linked list class

	Written by:	Richard Rodseth, Jens Alfke

	Copyright:	 1993-94 by Apple Computer, Inc., all rights reserved.

	Change History (most recent first):

		 <4>	 9/30/94	RR		#1167950. Added RemoveCurrent, fNext and
									fPrevious to FW_CPrivLinkedListIterator
		 <3>	 9/29/94	RA		1189812: Mods for 68K build.
		 <2>	  6/9/94	RR		Remove ASLM Stuff.
		 <3>	  5/9/94	jpa		Improvements for polygon clipper [1162090]
		 <2>	 3/15/94	MB		Changes to support SCpp/ASLM builds,
									#1150864.
		<11>	  2/7/94	JA		Capitalized THROW.
		<10>	  2/7/94	JA		Tiger Team Makeover. Added some inline
									helpers to eliminate cfront warnings in
									const methods.
		 <9>	 1/31/94	JA		Added more methods to FW_CPrivLink and FW_CPrivLinkedList.
									Inlined some getters/setters.
		 <7>	10/14/93	RR		Added fSeed and throw exception when
									iterator is out of sync
		 <6>	 8/13/93	RR		Fix up fSentinel in DeleteAllLinks
		 <5>	 8/13/93	RR		Added DeleteLinks method
		 <3>	 4/27/93	T		do some casting
		 <2>	  4/27/93	RCR		Include ODTypes and use FW_Boolean and NULL
		 <1>	  4/27/93	RCR		first checked in
	To Do:
*/

#include "FWFound.hpp"

#ifndef FWLIST_H
#include "FWList.h"
#endif

#ifndef FWDEBUG_H
#include "FWDebug.h"
#endif

#ifdef FW_BUILD_MAC
#include <Errors.h>
#endif

#if FW_LIB_EXPORT_PRAGMAS
#pragma lib_export on
#endif

#ifdef FW_BUILD_MAC
#pragma segment fwcollec
#endif

//==============================================================================
// Constants
//==============================================================================

//==============================================================================
// Scalar Types
//==============================================================================

//==============================================================================
// Local Classes
//==============================================================================

//==============================================================================
// Global Variables
//==============================================================================

//==============================================================================
// Function Prototype
//==============================================================================


//==============================================================================
// FW_CPrivLink
//==============================================================================



// Many of the simple link methods are inlines; see FWList.h for implementations.


//------------------------------------------------------------------------------
// FW_CPrivLink::FW_CPrivLink
//
// Constructor for FW_CPrivLink
//------------------------------------------------------------------------------

FW_CPrivLink::FW_CPrivLink()							
{ 
	fNext = NULL; 
	fPrevious = NULL;
}

//------------------------------------------------------------------------------
// FW_CPrivLink::FW_CPrivLink
//
// Constructor for FW_CPrivLink
//------------------------------------------------------------------------------

FW_CPrivLink::FW_CPrivLink(FW_CPrivLink* next, FW_CPrivLink* previous)							
{ 
	fNext = next; 
	fPrevious = previous;
}

//------------------------------------------------------------------------------
// FW_CPrivLink::FW_CPrivLink
//
// Copy constructor for FW_CPrivLink
//------------------------------------------------------------------------------

FW_CPrivLink::FW_CPrivLink( const FW_CPrivLink &link )							
{ 
	fNext = link.fNext; 
	fPrevious = link.fPrevious;
}

//------------------------------------------------------------------------------
// FW_CPrivLink::FW_CPrivLink
//
// Destructor for FW_CPrivLink
//------------------------------------------------------------------------------

FW_CPrivLink::~FW_CPrivLink()											
{
}

//------------------------------------------------------------------------------
// FW_CPrivLink::Remove
//
// Remove a link from its list (if any). DO NOT call this directly if there are
// any iterators active on the list; use FW_CPrivLinkedList::Remove instead.
//------------------------------------------------------------------------------

void	FW_CPrivLink::Remove( )
{
	if( fPrevious )
		fPrevious->SetNext(fNext);
	if( fNext )
		fNext->SetPrevious(fPrevious);
	fNext = NULL;
	fPrevious = NULL;
}

//------------------------------------------------------------------------------
// FW_CPrivLink::AddBefore
//
// Add a link to a list before another link. It must not already be on any list.
// DO NOT call this directly if there are any iterators active on the list;
// use FW_CPrivLinkedList::Remove instead.
//------------------------------------------------------------------------------

void	FW_CPrivLink::AddBefore( FW_CPrivLink *link )
{
	FW_ASSERT(link!=NULL);
	FW_ASSERT(fNext==NULL);
	FW_ASSERT(fPrevious==NULL);
	fNext = link;
	fPrevious = link->GetPrevious();
	fPrevious->SetNext(this);
	fNext->SetPrevious(this);
}

//------------------------------------------------------------------------------
// FW_CPrivLink::AddAfter
//
// Add a link to a list after another link. It must not already be on any list.
// DO NOT call this directly if there are any iterators active on the list;
// use FW_CPrivLinkedList::Remove instead.
//------------------------------------------------------------------------------

void	FW_CPrivLink::AddAfter( FW_CPrivLink *link )
{
	FW_ASSERT(link!=NULL);
	FW_ASSERT(fNext==NULL);
	FW_ASSERT(fPrevious==NULL);
	fPrevious = link;
	fNext = link->GetNext();
	fPrevious->SetNext(this);
	fNext->SetPrevious(this);
}

//======================================================================================
// Class FW_CPrivLinkedList
//======================================================================================

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::FW_CPrivLinkedList
//
// Constructor for FW_CPrivLinkedList
//------------------------------------------------------------------------------

FW_CPrivLinkedList::FW_CPrivLinkedList()
	:fSentinel(&fSentinel,&fSentinel),
	 fSeed(0)
{
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::~FW_CPrivLinkedList
//
// Destructor for FW_CPrivLinkedList
//------------------------------------------------------------------------------

FW_CPrivLinkedList::~FW_CPrivLinkedList()
{
	// The list does NOT delete all its links!
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::IsEmpty
//
// Description
//------------------------------------------------------------------------------

FW_Boolean FW_CPrivLinkedList::IsEmpty() const
{
	return this->IsSentinel( fSentinel.GetNext() );
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::Count
//
// Description
//------------------------------------------------------------------------------

unsigned long FW_CPrivLinkedList::Count() const
{
	FW_CPrivLink* l = fSentinel.GetNext();
	unsigned long count = 0;
	while (this->NotSentinel(l))
	{
		count++;
		l = l->GetNext();
	}
	return count;
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::Includes
//
// Does the list include a particular link?
//------------------------------------------------------------------------------

FW_Boolean FW_CPrivLinkedList::Includes( const FW_CPrivLink *link ) const
{
	FW_CPrivLink* l = fSentinel.GetNext();
	while (this->NotSentinel(l))
	{
		if( l == link )
			return TRUE;
		else
			l = l->GetNext();
	}
	return FALSE;
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::DeleteAllLinks
//
// Description
//------------------------------------------------------------------------------

void	FW_CPrivLinkedList::DeleteAllLinks()
{

	FW_CPrivLink* l = fSentinel.GetNext();
	FW_CPrivLink* n = l;
	while (this->NotSentinel(n))
	{
		n = l->GetNext();
		delete l;	
		l = n;
	}
	fSentinel.SetNext(&fSentinel);
	fSentinel.SetPrevious(&fSentinel);
	fSeed++;
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::RemoveAll
//
// Description
//------------------------------------------------------------------------------

void	FW_CPrivLinkedList::RemoveAll()
{

	FW_CPrivLink* l = fSentinel.GetNext();
	FW_CPrivLink* n = l;
	while (this->NotSentinel(l))
	{
		n = l->GetNext();
		l->SetNext(NULL);
		l->SetPrevious(NULL);
		l = n;
	}
	fSentinel.SetNext(&fSentinel);
	fSentinel.SetPrevious(&fSentinel);
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::Remove
//
// Description
//------------------------------------------------------------------------------

void	FW_CPrivLinkedList::Remove(FW_CPrivLink& aLink)
{
	fSeed++;
	aLink.Remove();
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::RemoveFirst
//
// Description
//------------------------------------------------------------------------------

FW_CPrivLink* FW_CPrivLinkedList::RemoveFirst()
{
	FW_CPrivLink* old = fSentinel.GetNext();
	if (this->NotSentinel(old))
	{
		fSeed++;
		old->Remove();
		return old;
	}
	else
	{
		return NULL;
	}
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::RemoveLast
//
// Description
//------------------------------------------------------------------------------

FW_CPrivLink* FW_CPrivLinkedList::RemoveLast()
{
	FW_CPrivLink* old = fSentinel.GetPrevious();
	if (this->NotSentinel(old))
	{
		fSeed++;
		old->Remove();
		return old;
	}
	else
	{
		return NULL;
	}
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::AddFirst
//
// Description
//------------------------------------------------------------------------------

void	FW_CPrivLinkedList::AddFirst(FW_CPrivLink* link)
{
	link->AddAfter(this->GetSentinel());
	fSeed++;
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::AddLast( FW_CPrivLink )
//
// Description
//------------------------------------------------------------------------------

void	FW_CPrivLinkedList::AddLast(FW_CPrivLink* link)
{
	link->AddBefore(this->GetSentinel());
	fSeed++;
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::AddLast( FW_CPrivLinkedList )
//
// Append contents of another list to my end. (Other list becomes empty.)
//------------------------------------------------------------------------------

void	FW_CPrivLinkedList::AddLast( FW_CPrivLinkedList &list )
{
	if( !list.IsEmpty() ) {
		FW_CPrivLink *myLast = fSentinel.GetPrevious();
		FW_CPrivLink *itsFirst = list.First();
		myLast->SetNext(itsFirst);
		itsFirst->SetPrevious(myLast);
		
		FW_CPrivLink *itsLast = list.Last();
		itsLast->SetNext(this->GetSentinel());
		fSentinel.SetPrevious(itsLast);
		
		list.fSentinel.SetNext(this->GetSentinel());
		list.fSentinel.SetPrevious(this->GetSentinel());

		fSeed++;
		list.fSeed++;
	}
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::AddLastUnique( FW_CPrivLinkedList )
//
// Append contents of another list (w/o duplication) to my end.
//------------------------------------------------------------------------------

void	FW_CPrivLinkedList::AddLastUnique( FW_CPrivLinkedList &list )
{
	for( FW_CPrivLink *link = fSentinel.GetNext(); this->NotSentinel(link); link=link->GetNext() )
		if( list.Includes(link) )
			list.Remove(*link);
	this->AddLast(list);
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::AddBefore
//
// Description
//------------------------------------------------------------------------------

void FW_CPrivLinkedList::AddBefore(FW_CPrivLink& existing, FW_CPrivLink* link)
{
	link->AddBefore(&existing);
	fSeed++;
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::AddAfter
//
// Description
//------------------------------------------------------------------------------

void FW_CPrivLinkedList::AddAfter(FW_CPrivLink& existing, FW_CPrivLink* link)
{
	link->AddAfter(&existing);
	fSeed++;
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::After
//
// Description
//------------------------------------------------------------------------------

FW_CPrivLink* FW_CPrivLinkedList::After(const FW_CPrivLink& link) const
{
	FW_CPrivLink *next = link.GetNext();
	if( this->IsSentinel(next) )
		return NULL;
	else
		return next;
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::Before
//
// Description
//------------------------------------------------------------------------------

FW_CPrivLink* FW_CPrivLinkedList::Before(const FW_CPrivLink& link) const
{
	FW_CPrivLink *prev = link.GetPrevious();
	if( this->IsSentinel(prev) )
		return NULL;
	else
		return prev;
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::First
//
// Description
//------------------------------------------------------------------------------

FW_CPrivLink*	FW_CPrivLinkedList::First()  const
{
	return this->NotSentinel(fSentinel.GetNext()) ? fSentinel.GetNext() : (FW_CPrivLink*) NULL;
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedList::Last
//
// Description
//------------------------------------------------------------------------------

FW_CPrivLink* FW_CPrivLinkedList::Last() const
{
	return this->NotSentinel(fSentinel.GetPrevious()) ? fSentinel.GetPrevious() : (FW_CPrivLink*) NULL;
}


//======================================================================================
// Class FW_CPrivLinkedListIterator
//======================================================================================

//------------------------------------------------------------------------------
// FW_CPrivLinkedListIterator::FW_CPrivLinkedListIterator
//
// Constructor for FW_CPrivLinkedListIterator
//------------------------------------------------------------------------------

FW_CPrivLinkedListIterator::FW_CPrivLinkedListIterator(FW_CPrivLinkedList* list)
{
	fList = list;
	fCurrent = NULL;
	fNext = NULL;
	fPrevious = NULL;
	fSentinel = &list->fSentinel;
	fSeed = fList->fSeed;	
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedListIterator::~FW_CPrivLinkedListIterator
//
// Destructor for FW_CPrivLinkedListIterator
//------------------------------------------------------------------------------

FW_CPrivLinkedListIterator::~FW_CPrivLinkedListIterator()
{
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedListIterator::First
//
// Description
//------------------------------------------------------------------------------

FW_CPrivLink* FW_CPrivLinkedListIterator::First()
{
	FW_ASSERT(fSeed == fList->fSeed);
		
	fCurrent = fList->First();
	if (fCurrent == fSentinel)
		fCurrent = NULL;
	return fCurrent;
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedListIterator::Next
//
// Description
//------------------------------------------------------------------------------

FW_CPrivLink* FW_CPrivLinkedListIterator::Next()
{
	FW_ASSERT(fSeed == fList->fSeed);

	if (fCurrent == NULL)
	{
		if ((fNext == NULL) && (fPrevious == NULL))	// Just starting out
		{
			return this->First();
		}
		else	// Just deleted
		{
			fCurrent = fNext;
			fPrevious = NULL;
			fNext = NULL;
		}
	}
	else
		fCurrent = fCurrent->GetNext();
	
	if (fCurrent == fSentinel)
		fCurrent = NULL;
	return fCurrent;
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedListIterator::Last
//
// Description
//------------------------------------------------------------------------------

FW_CPrivLink* FW_CPrivLinkedListIterator::Last()
{
	FW_ASSERT(fSeed == fList->fSeed);

	fCurrent = fList->Last();
	if (fCurrent == fSentinel)
		fCurrent = NULL;
	return fCurrent;
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedListIterator::Previous
//
// Description
//------------------------------------------------------------------------------

FW_CPrivLink* FW_CPrivLinkedListIterator::Previous()
{
	FW_ASSERT(fSeed == fList->fSeed);

	if (fCurrent == NULL)
	{
		if ((fNext == NULL) && (fPrevious == NULL))	// Just starting out
		{
			return this->Last();
		}
		else	// Just deleted
		{
			fCurrent = fPrevious;
			fPrevious = NULL;
			fNext = NULL;
		}
	}
	else
		fCurrent = fCurrent->GetPrevious();

	if (fCurrent == fSentinel)
		fCurrent = NULL;
	return fCurrent;
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedListIterator::Current
//
// Description
//------------------------------------------------------------------------------

FW_CPrivLink* FW_CPrivLinkedListIterator::Current()
{
	return fCurrent;
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedListIterator::IsNotComplete
//
// Description
//------------------------------------------------------------------------------

FW_Boolean FW_CPrivLinkedListIterator::IsNotComplete()
{
	return (fCurrent != NULL);
}

//------------------------------------------------------------------------------
// FW_CPrivLinkedListIterator::RemoveCurrent
//
// Description
//------------------------------------------------------------------------------

void  FW_CPrivLinkedListIterator::RemoveCurrent()
{
	FW_ASSERT(fSeed == fList->fSeed);
	 	
	if (fCurrent != NULL)
	{
		fNext = fCurrent->GetNext();
		fPrevious = fCurrent->GetPrevious();
			
		fList->Remove(*fCurrent);
		fCurrent = NULL;
		fSeed = fList->fSeed;
	}
}
